﻿namespace sample
open System
open System.Collections.Generic
open System.Windows.Forms
open System.Threading
open System.Drawing
open System.Drawing.Imaging
open OpenTK
open OpenTK.Graphics
open OpenTK.Graphics.OpenGL
open System
open System.Collections.Generic
open System.Windows.Forms
open System.Threading
open System.Drawing
open System.Drawing.Imaging
open OpenTK
open OpenTK.Graphics
open OpenTK.Graphics.OpenGL

module Shape =

    type T2N3V3 =
        struct
            val mutable TexCoord : Vector2
            val mutable Normal : Vector3
            val mutable Position : Vector3
            new(v,n,p) = {TexCoord = v;Normal = n;Position = p}
        end

    [<AbstractClass>]
    type Shape() =
        let mutable bCreate = false
        let mutable vi = 0
        let mutable ei = 0
        let mutable count = 0
        member this.vboID with get() = vi and set value = vi <- value
        member this.eboID with get() = ei and set value = ei <- value
        member this.TriangelCount with get() = count and set value = count <- value
        member this.IsCreate with get() = bCreate and set value = bCreate <- value
        abstract Draw : unit -> unit
        abstract Init : unit -> unit
        member this.InitQ : unit -> unit =fun () -> ()

    type Sphere(radius:float32,level:int) =
        inherit Shape()
        let mutable rad,lev = radius,level
        let RightLevel = 
            if lev < 0 then lev <- 0
            elif lev > 6 then lev <-6
        override this.Draw() =  
            if this.IsCreate<>true then this.Init()
            GL.BindBuffer(BufferTarget.ArrayBuffer,this.vboID)
            GL.BindBuffer(BufferTarget.ElementArrayBuffer,this.eboID)
            GL.InterleavedArrays(InterleavedArrayFormat.T2fN3fV3f,0,IntPtr.Zero)
            GL.DrawElements(BeginMode.Triangles,this.TriangelCount,DrawElementsType.UnsignedInt,IntPtr.Zero)
        override this.Init() =
            let alls = Array.create 6 (new T2N3V3())
            alls.[0] <- new T2N3V3(new Vector2( 0.0f, 0.0f ), Vector3.UnitX, Vector3.UnitX * rad )
            alls.[1] <- new T2N3V3(new Vector2( 0.0f, 0.0f ), Vector3.UnitY, Vector3.UnitY * rad )
            alls.[2] <- new T2N3V3(new Vector2( 0.0f, 0.0f ), Vector3.UnitZ, Vector3.UnitZ * rad )
            alls.[3] <- new T2N3V3(new Vector2( 0.0f, 0.0f ), -Vector3.UnitX, -Vector3.UnitX * rad )
            alls.[4] <- new T2N3V3(new Vector2( 0.0f, 0.0f ), -Vector3.UnitY, -Vector3.UnitY * rad )
            alls.[5] <- new T2N3V3(new Vector2( 0.0f, 0.0f ), -Vector3.UnitZ, -Vector3.UnitZ * rad )
            let is = [| 
                            1;2;0
                            0;2;4
                            0;4;5
                            5;1;0
                            1;3;2
                            4;2;3
                            4;3;5
                            1;5;3
                        |]
            let (vvv:T2N3V3 []),(iv: int[]) = this.Sub (alls,is)
            let mutable vID,eID = 0,0
            //let mutable tv,vv,pv = vvv |> Array.map (fun v -> v.TexCoord,v.Normal,v.Position) |> Array.unzip3            
            GL.GenBuffers(1,&vID) 
            GL.BindBuffer(BufferTarget.ArrayBuffer,vID)
            GL.BufferData(BufferTarget.ArrayBuffer,IntPtr (4 * 8 * vvv.Length),vvv,BufferUsageHint.StaticDraw)
            
            GL.GenBuffers(1,&eID)
            GL.BindBuffer(BufferTarget.ElementArrayBuffer,eID)
            GL.BufferData(BufferTarget.ElementArrayBuffer,IntPtr (4 * iv.Length),iv,BufferUsageHint.StaticDraw)

            this.vboID <- vID
            this.eboID <- eID
            this.TriangelCount <- iv.Length 
            this.IsCreate <- true
            ()
        member v.GetMidValue (first:T2N3V3,second:T2N3V3) =
            let midN = Vector3.Lerp(first.Position,second.Position,0.5f) |> Vector3.Normalize
            let midP = midN *(float32 rad)
            let midT = Vector2.Lerp(first.TexCoord,second.TexCoord,0.5f) |> Vector2.Normalize
            let result = new T2N3V3(midT,midN,midP)
            result
        member v.Subdivide (v1:T2N3V3,v2:T2N3V3,v3:T2N3V3) =
            let vs = Array.create 6 (new T2N3V3())
            vs.[0] <- v1
            vs.[1] <- v.GetMidValue(v1,v2)
            vs.[2] <- v.GetMidValue(v3,v1)
            vs.[3] <- v2
            vs.[4] <- v.GetMidValue(v2,v3)
            vs.[5] <- v3
            let is = Array.create 12 0
            is.[0] <- 0
            is.[1] <- 1
            is.[2] <- 2
            is.[3] <- 2
            is.[4] <- 1
            is.[5] <- 4
            is.[6] <- 4
            is.[7] <- 1
            is.[8] <- 3
            is.[9] <- 2
            is.[10] <-4
            is.[11] <- 5
            (vs,is)
        member this.Sub(alls:T2N3V3 [],is:int []) =
            //let mutable tv,vv,pv = alls |> Array.map (fun v -> v.TexCoord,v.Normal,v.Position) |> Array.unzip3
            let mutable allv = alls
            let mutable iv = is
            let show array = printfn "%A" array
            for j in 0 .. lev do
                let mutable av = Array.create 0 (new T2N3V3())
                let mutable ev = Array.create 0 0
                printfn "%i" allv.Length
                printfn "%i" iv.Length
                for i in 0 .. 3 .. iv.Length - 1 do
                    let (vvv,iiv) = this.Subdivide(allv.[iv.[i]],allv.[iv.[i+1]],allv.[iv.[i+2]])
                    let length = av.Length
                    av <- Array.append av vvv
                    let map = iiv |> Array.map (fun p -> p + length)
                    ev <- Array.append ev map
                allv <- av
                iv <- ev 
                allv |> Array.map (fun p -> p.Position) |> show
                show iv
            allv,iv

    type Cube(width:float32,height:float32,length:float32,index:int) =
        inherit Shape()
        let mutable id = index
        let xl,yl,zl =width/2.f,height/2.f,length/2.f
        let mutable color = Color.White
        let drawTexture index texID = ()
        let v8 = [|
                        new Vector3(xl,yl,zl)             //0
                        new Vector3(-xl,yl,zl)            //1
                        new Vector3(-xl,-yl,zl)          //2
                        new Vector3(xl,-yl,zl)            //3
                        new Vector3(xl,yl,-zl)            //4
                        new Vector3(-xl,yl,-zl)          //5
                        new Vector3(-xl,-yl,-zl)         //6
                        new Vector3(xl,-yl,-zl)          //7
                    |]
        let eArray = [|
                        0;1;2;0;2;3
                        4;5;1;4;1;0
                        4;0;3;4;3;7
                        3;2;6;3;6;7
                        1;5;6;1;6;2
                        5;4;7;5;7;6
                    |]        
        new(x,y,z) =
            let rnd = System.Random().Next()
            printfn "%i" rnd
            Cube(x,y,z,-1)
        override this.Init() =
            let mutable vID = 0
            GL.GenBuffers(1,&vID) 
            GL.BindBuffer(BufferTarget.ArrayBuffer,vID)
            GL.BufferData(BufferTarget.ArrayBuffer,IntPtr (4 * 3 * v8.Length),v8,BufferUsageHint.StaticDraw)
            this.vboID <- vID
            printfn "vID:%i" vID
            let mutable eID = 0
            GL.GenBuffers(1,&eID)
            GL.BindBuffer(BufferTarget.ElementArrayBuffer,eID)
            GL.BufferData(BufferTarget.ElementArrayBuffer,IntPtr (4 * eArray.Length),eArray,BufferUsageHint.StaticDraw)
            this.eboID <- eID
            this.IsCreate <- true    
            let rnd = System.Random(this.GetHashCode())
            this.Color <- Color.FromArgb(rnd.Next(0,255),rnd.Next(0,255),rnd.Next(0,255))
            ()
        member this.Index with get() = id and set value = id <-value
        member this.Color with get() = color and set value = color <- value
        override this.Draw() = 
            this.DrawTexTure(0,0,false)
        member this.DrawTexTure(index:int,texId:int,?bAll:bool) =
            let ball = defaultArg bAll false
            let ts = Array.create 1 (index,texId,ball)
            this.DrawTexTure(ts)
        member this.DrawTexTure(its:(int*int*bool)[]) = 
            if this.IsCreate<>true then this.Init()            
            GL.BindBuffer(BufferTarget.ArrayBuffer,this.vboID)
            GL.BindBuffer(BufferTarget.ElementArrayBuffer,this.eboID)
            GL.InterleavedArrays(InterleavedArrayFormat.V3f,0,IntPtr.Zero)
            //GL.VertexPointer(3,VertexPointerType.Float,0,IntPtr.Zero)
            GL.PushMatrix()
            if id >= 0 && id < 8 then
                GL.Translate(v8.[id])
            let getti index =
                let result = Array.tryFind (fun p -> 
                                                        let x,_,_ = p
                                                        x = index)  its
                match result with
                | Some x -> x
                | None -> 0,0,false
            this.DrawTexTure(getti(0),Vector3.UnitZ,[|0;1;2;0;2;3|])
            this.DrawTexTure(getti(1),Vector3.UnitY,[|4;5;1;4;1;0|])
            this.DrawTexTure(getti(2),Vector3.UnitX,[|4;0;3;4;3;7|])
            this.DrawTexTure(getti(3),-Vector3.UnitY,[|3;2;6;3;6;7|])
            this.DrawTexTure(getti(4),-Vector3.UnitX,[|1;5;6;1;6;2|])
            this.DrawTexTure(getti(5),-Vector3.UnitZ,[|5;4;7;5;7;6|])
            GL.PopMatrix()
        member this.DrawTexTure(ts:int*int*bool,vector:Vector3,indexs:int []) =
            let ind,txtId,ball = ts
            if txtId <> 0  then 
                GL.BindTexture(TextureTarget.Texture2D,txtId)
                this.GenTexture(vector,ball)
            GL.Normal3(vector)
            GL.DrawElements(BeginMode.Triangles,6,DrawElementsType.UnsignedInt,IntPtr (ind * 24))//,indexs)
            GL.Disable(EnableCap.TextureGenS)
            GL.Disable(EnableCap.TextureGenT)
            GL.Disable(EnableCap.TextureGenR)
            GL.Disable(EnableCap.TextureGenQ)

        member this.GenTexture(vector:Vector3,ball) = 
            GL.Enable(EnableCap.TextureGenS)
            GL.Enable(EnableCap.TextureGenT)
            GL.TexGen(TextureCoordName.S,TextureGenParameter.TextureGenMode,int TextureGenMode.ObjectLinear)
            GL.TexGen(TextureCoordName.T,TextureGenParameter.TextureGenMode,int TextureGenMode.ObjectLinear)
            let mutable xa = [|1.f;0.f;0.f;0.f|]
            let mutable xy = [|0.f;1.f;0.f;0.f|]
            let mutable x,y,z=1.f,1.f,0.f
            if ball then z <-0.5f
            if vector = Vector3.UnitY || vector = -Vector3.UnitY then 
                if ball then
                    x<-1.0f/Vector3.Subtract(v8.[0],v8.[1]).Length
                    y<-1.0f/Vector3.Subtract(v8.[0],v8.[4]).Length
                xa <- [|x;0.f;0.f;z|]
                xy <- [|0.f;0.f;y;z|]
            if vector = Vector3.UnitZ || vector = -Vector3.UnitZ then 
                if ball then
                    x<-1.0f/Vector3.Subtract(v8.[0],v8.[1]).Length
                    y<-1.0f/Vector3.Subtract(v8.[0],v8.[3]).Length
                xa <- [|x;0.f;0.f;z|]
                xy <- [|0.f;y;0.f;z|]
            if vector = Vector3.UnitX || vector = -Vector3.UnitX then 
                if ball then                
                    x<-1.0f/Vector3.Subtract(v8.[0],v8.[4]).Length
                    y<-1.0f/Vector3.Subtract(v8.[0],v8.[3]).Length
                xa <- [|0.f;0.f;x;z|]
                xy <- [|0.f;y;0.f;z|]
            GL.TexGen(TextureCoordName.S,TextureGenParameter.ObjectPlane,xa)
            GL.TexGen(TextureCoordName.T,TextureGenParameter.ObjectPlane,xy)

    type Camera() = 
        let mutable origin = Vector3.Zero
        let mutable length = 1.
        let mutable yangle = 0.
        let mutable xangle= Math.PI/2.
        let mutable bThree = true
        do
            if not bThree then xangle <- Math.PI/2. else xangle <- 1.5*Math.PI            
        member this.IsThree 
            with get() = bThree 
            and set value = 
                //if value then this.Origin <- this.Target
                xangle <- xangle + Math.PI
                bThree <- value
        member this.Eye 
            with get() = 
                let mutable eye = this.Origin
                if bThree then eye <- this.Direction
                eye
        member this.People 
            with get() = 
                let mutable people = this.Origin
                if not bThree then people <- this.Direction
                people
        member this.Target 
             with get() = 
                let mutable target = this.Direction
                if bThree then target <- this.Origin
                target
        member this.Origin 
            with get() = origin 
            and set value = origin <- value
        member this.Length 
            with get() = length 
            and set value = 
                if value < 0. then length <- 0.1
                length <- value
        member this.YAngle 
            with get() = yangle
            and set value = 
                if value > Math.PI/2. then yangle <- Math.PI/2.
                elif value < -Math.PI/2. then yangle <- -Math.PI/2.
                else yangle <- value
        member this.XAngle 
            with get() = xangle
            and set value = 
                if value > 2.* Math.PI then xangle <- value - 2.* Math.PI
                elif value < 0. then xangle <- value + 2. * Math.PI
                else xangle <- value       
        member this.PeopleAngle 
            with get()=
                let mutable angle = this.XAngle + Math.PI/2.
                angle
        member this.Direction 
            with get() = 
                let mutable len = 1.
                if bThree then len <- length
                let xyLength = Math.Cos(this.YAngle) * len
                let x:float =float origin.X + xyLength * Math.Cos(this.XAngle)
                let y:float =float origin.Y + len * Math.Sin(this.YAngle)
                let z:float =float origin.Z + xyLength * Math.Sin(this.XAngle)
                Vector3(float32 x,float32 y,float32 z)
        member this.Transelt (x,y,z) = 
            let sinX = Math.Sin(this.XAngle)
            let cosX = Math.Cos(this.XAngle)
            let mutable xstep = x * sinX + z * cosX 
            let mutable zstep = z * sinX - x * cosX
            if bThree then 
                xstep <- -xstep
                zstep <- -zstep
            let x1 = float origin.X + xstep
            let y1 = float origin.Y + y
            let z1 = float origin.Z + zstep 
            printfn "angle:%f, sinx:%f, cosx:%f" this.XAngle sinX cosX
            printfn "x:%f, y:%f, z:%f" x1 y1 z1
            origin <- new Vector3(float32 x1,float32 y1,float32 z1)
        member this.Rotate (x,y) =
            let xa = this.XAngle + x
            let mutable ystep = y
            if bThree then ystep <- -y
            let ya =this.YAngle + ystep
            this.YAngle <- ya
            this.XAngle <- xa

    type HLight(index:LightName) =
        //opengl的光源编号
        member val Index = index
        //光源所在位置
        member val Location = [|0.f;0.f;0.f;0.f|] with get, set
        //光源的环境强度
        member val Ambient = [|0.f;0.f;0.f;0.f|] with get, set
        //光源的散射强度
        member val Diffuse = [|0.f;0.f;0.f;0.f|] with get, set
        //光源的境面强度
        member val Specular = [|0.f;0.f;0.f;0.f|] with get, set    
        //是否是聚光灯
        member val IsSpot = false with get,set
        //聚光灯方向
        member val SpotDirection = [|0.f;0.f;0.f;0.f|] with get, set   
        //聚光灯的切角
        member val SpotCutoff = 90.f with get, set 
        //是否打开
        member val IsOpen = false with get,set
        member this.Light() = 
            let enableCap = enum<EnableCap>(int this.Index)
            if this.IsOpen then
                GL.Enable(enableCap)
                GL.Light(this.Index,LightParameter.Ambient,this.Ambient)
                GL.Light(this.Index,LightParameter.Diffuse,this.Diffuse)
                GL.Light(this.Index,LightParameter.Position,this.Location)
                GL.Light(this.Index,LightParameter.Specular,this.Specular)
                if this.IsSpot then
                    GL.Light(this.Index,LightParameter.SpotDirection,this.SpotDirection)
                    GL.Light(this.Index,LightParameter.SpotCutoff,this.SpotCutoff)
            else
               GL.Disable(enableCap) 
            
        member this.Copy (index:LightName) = 
            let light = new HLight(index)
            light.Location <- this.Location
            light.Ambient <- this.Ambient
            light.Diffuse <- this.Diffuse
            light.Specular <- this.Specular
            light.IsSpot <- this.IsSpot
            light.IsOpen <- this.IsOpen
            light.SpotDirection <- this.SpotDirection
            light.SpotCutoff <- this.SpotCutoff
            light
            
type TexTure() =
    static member Load (path:string) =
        let mutable texture = 0
        GL.GenTextures(1,&texture)
        let bitmap = new Bitmap(path)
        GL.BindTexture(TextureTarget.Texture2D,texture)
        let data = bitmap.LockBits(new System.Drawing.Rectangle(0,0,bitmap.Width,bitmap.Height),ImageLockMode.ReadOnly,System.Drawing.Imaging.PixelFormat.Format32bppArgb)
        GL.TexImage2D(TextureTarget.Texture2D,0,PixelInternalFormat.Rgba, data.Width, data.Height, 0,
            OpenTK.Graphics.OpenGL.PixelFormat.Bgra, PixelType.UnsignedByte, data.Scan0)
        bitmap.UnlockBits(data)
        GL.TexParameter(TextureTarget.Texture2D,TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear)
        GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear)
        texture
    static member Load(path:string []) =
        let mutable texs = Array.create path.Length 0
        let len = path.Length - 1
        for i=0 to len do
                GL.GenTextures(1,&texs.[i])
                let bitmap = new Bitmap(path.[i])
                GL.BindTexture(TextureTarget.Texture2D,texs.[i])
                let data = bitmap.LockBits(new System.Drawing.Rectangle(0,0,bitmap.Width,bitmap.Height),ImageLockMode.ReadOnly,System.Drawing.Imaging.PixelFormat.Format32bppArgb)
                GL.TexImage2D(TextureTarget.Texture2D,0,PixelInternalFormat.Rgba, data.Width, data.Height, 0,
                    OpenTK.Graphics.OpenGL.PixelFormat.Bgra, PixelType.UnsignedByte, data.Scan0)
                bitmap.UnlockBits(data)
                GL.TexParameter(TextureTarget.Texture2D,TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear)
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear)              
        texs
 
